package xlogo.kernel;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.event.AdjustmentEvent;
import java.awt.geom.Arc2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.Stack;
import java.util.StringTokenizer;
import java.awt.event.*;
import xlogo.Application;
import xlogo.gui.Ardoise;
import xlogo.gui.preferences.Panel_Font;
import xlogo.utils.Utils;
import xlogo.utils.myException;
import xlogo.Config;
import xlogo.Logo;

/**
 * Title :        XLogo
 * Description :  XLogo is an interpreter for the Logo 
 * 						programming language
 * @author Loïc Le Coq
 */public class DrawGraphics implements AdjustmentListener{
	
	public Turtle tortue;
	public Turtle[] tortues; 
	public Stack tortues_visibles;
	private Kernel kernel; 

	//0 si mode fenetre 1 si mode enroule 2 si mode clos
	protected static int etat_fenetre = 0;
	protected static final int WINDOW_CLASSIC=0;
	protected static final int WINDOW_WRAP=1;
	protected static final int WINDOW_CLOSE=2;
	
	// Boolean for animation mode
	private boolean classicMode=true; // true si classique false si animation
	private final boolean MODE_ANIMATION=false;
	private final boolean MODE_CLASSIC=true;;

	// default predefined colors 
	public static final Color[] defaultColors={Color.BLACK,Color.RED,Color.GREEN,Color.YELLOW,Color.BLUE,
			Color.MAGENTA,Color.CYAN,Color.WHITE,Color.GRAY,Color.LIGHT_GRAY,new Color(128,0,0),new Color(0,128,0),
			new Color(0,0,128),Color.ORANGE,Color.PINK,new Color(128,0,255),new Color(153,102,0)};
	// numéro identifiant la police de la
	protected int police_etiquette;
	private Color couleurfond = Color.white;
	
	private GeneralPath gp;
	private Arc2D arc,clipArc;
	private Line2D line; 
	private Rectangle2D rec;
	private Application cadre;
	private Graphics graph;
	private Graphics2D g;
	private BasicStroke crayon_triangle = new BasicStroke(1);
	private double[] coords;
	
	
	private double oldx, oldy,x1,y1,x2,y2,nx,ny,vx,vy,factor,length;
	protected DrawGraphics(Application cadre){
		this.cadre=cadre;
		this.kernel=cadre.getKernel();
		police_etiquette=Application.police;
		// init all turtles
		tortues = new Turtle[Config.maxTurtles];
		tortues_visibles=new Stack();
		tortue=new Turtle(cadre);
		tortue.id = 0;
		tortues_visibles.push(String.valueOf("0"));
		for (int i = 0; i < tortues.length; i++) { //initialisation de
			// l'ensemble des tortues
			if (i == 0)
				tortues[i] = tortue;
			else
				tortues[i] = null;
		}
		g = (Graphics2D) Ardoise.dessin.getGraphics();
		graph = cadre.getArdoise().getGraphics();
		// Ajouter les listeners au scrollbar du cadre
		cadre.jScrollPane1.getHorizontalScrollBar().addAdjustmentListener(this);
		cadre.jScrollPane1.getVerticalScrollBar().addAdjustmentListener(this);
		couleurfond=Config.screencolor;
		this.setQuality(Config.quality);
		init();
	}
	protected void av(double arg) {
		oldx = tortue.corX;
		oldy = tortue.corY;
		if (DrawGraphics.etat_fenetre == DrawGraphics.WINDOW_CLASSIC) { //mode fenetre
					montrecacheTortue();
			tortue.corX = tortue.corX + arg
					* Math.cos(tortue.angle);
			tortue.corY = tortue.corY - arg
					* Math.sin(tortue.angle);
			if (tortue.isPenDown()) {
				if (tortue.isPenReverse()) {
					g.setColor(couleurfond);
					g.setXORMode(tortue.couleurcrayon);
				} else {
					g.setColor(tortue.couleurcrayon);
					g.setPaintMode();
				}
				if (null==line) line=new Line2D.Double();
				if (null==gp) gp=new GeneralPath();
				else gp.reset();
				if (oldx < tortue.corX){
					x1=oldx;y1=oldy;x2=tortue.corX;y2=tortue.corY;
				}
				if (oldx>tortue.corX){
					x2=oldx;y2=oldy;x1=tortue.corX;y1=tortue.corY;
				}
				else if (oldx==tortue.corX){
					if (oldy<tortue.corY){
						x2=oldx;y2=oldy;x1=tortue.corX;y1=tortue.corY;						
						}
					else{ 
						x1=oldx;y1=oldy;x2=tortue.corX;y2=tortue.corY;
					}
				}
				line.setLine(x1,y1,x2,y2);
				// perpendicular vector
				nx=y1-y2;
				ny=x2-x1;
				length=Math.sqrt(nx*nx+ny*ny);
				if (length!=0){
					factor=(1+tortue.getPenWidth())/length;
					vx=x2-x1;
					vy=y2-y1;
					gp.moveTo((float)(x1-vx*factor-nx*factor),
						(float)(y1-vy*factor-ny*factor));
					gp.lineTo((float)(x1-vx*factor+nx*factor),
						(float)(y1-vy*factor+ny*factor));		
					gp.lineTo((float)(x2+vx*factor+nx*factor),
						(float)(y2+vy*factor+ny*factor));
					gp.lineTo((float)(x2+vx*factor-nx*factor),
						(float)(y2+vy*factor-ny*factor));
					gp.lineTo((float)(x1-vx*factor-nx*factor),
						(float)(y1-vy*factor-ny*factor));	
				}
				else{
					float width=tortue.getPenWidth()+0.5f;
					gp.moveTo((float)(x1-width),
							(float)(y1-width));
						gp.lineTo((float)(x1+width),
							(float)(y1-width));		
						gp.lineTo((float)(x1+width),
							(float)(y1+width));
						gp.lineTo((float)(x1-width),
							(float)(y1+width));
						gp.lineTo((float)(x1-width),
							(float)(y1-width));
				}
		/*		if (tortue.isPenReverse()){
					if (oldx>0 && oldy>0 && oldx<Config.imageHeight && oldy < Config.imageWidth) {
						if (rec==null) rec=new Rectangle2D.Double();
						 // High quality
						if (Config.quality==Config.quality_HIGH){
							double width=tortue.getPenWidth();
							rec.setRect(oldx-width,oldy-width,
									2*width,2*width);
						}
						// Normal or Low Quality
						else{
							// penWidth is 2k or 2k+1??
							int intWidth=(int)(2*tortue.getPenWidth()+0.5);
							if (intWidth%2==1){
//								System.out.println("coucou");
								double width=tortue.getPenWidth()-0.5;
	//							System.out.println(width+" "+oldx+" "+oldy);
								
								rec.setRect(oldx-width,oldy-width,
										2*width+1,2*width+1);
							}
							else {
								double width=tortue.getPenWidth();
								rec.setRect(oldx-width,oldy-width,
										2*width,2*width);	
							}
						}
							if (Config.penShape==Logo.PEN_SHAPE_SQUARE){
								//System.out.println(rec.getBounds().toString()+" "+tortue.angle);
								//AffineTransform at=AffineTransform.getRotateInstance(tortue.angle-Math.PI/2,oldx+0.5,oldy+0.5);
								
								//Shape rec2=at.createTransformedShape(rec);
//								System.out.println(rec2.getBounds().toString());
								g.fill(rec);
							}
							else if (Config.penShape==Logo.PEN_SHAPE_OVAL){
								if (null==arc) arc=new Arc2D.Double();
								arc.setArcByCenter(x2,y2,0,0,360,Arc2D.OPEN);
								g.draw(arc);
							}
							clip(rec);

						}
	
				
				
				
				}*/
				g.draw(line);
				clip(gp);

			}
			montrecacheTortue();
		} else if (DrawGraphics.etat_fenetre == DrawGraphics.WINDOW_WRAP) { //mode enroule
			trace_enroule(arg, oldx, oldy);
		} else if (DrawGraphics.etat_fenetre == DrawGraphics.WINDOW_CLOSE) { //mode clos
			try {
				trace_ferme(oldx, oldy, arg);
			} catch (myException e) {
			}
		}
	}

	protected void arc(double rayon, double pangle, double fangle) {
		// Put fangle and pangle between 0 and 360
		fangle = ((90 - fangle) % 360); 
		pangle = ((90 - pangle) % 360);
		if (fangle<0) fangle+=360;
		if (pangle<0) pangle+=360;
		// Calculate angle extend
		double angle=pangle-fangle;
		if (angle<0) angle+=360;
		montrecacheTortue();
		if (null==arc) arc=new Arc2D.Double();
		arc.setArcByCenter(tortue.corX,tortue.corY,rayon,
				fangle,angle, Arc2D.OPEN);
		if (tortue.isPenReverse()) {
			g.setColor(couleurfond);
			g.setXORMode(tortue.couleurcrayon);
		} else {
			g.setColor(tortue.couleurcrayon);
			g.setPaintMode();
		}

		g.draw(arc);
/*		if (null==gp) gp=new GeneralPath();
		else gp.reset();
		gp.moveTo((float)(tortue.corX-rayon-tortue.taille_crayon),
				(float)(tortue.corY-rayon-tortue.taille_crayon));
		gp.lineTo((float)(tortue.corX-rayon-tortue.taille_crayon),
				(float)(tortue.corY-rayon-tortue.taille_crayon));
		gp.lineTo((float)(tortue.corX-rayon-tortue.taille_crayon),
				(float)(tortue.corY-rayon-tortue.taille_crayon));
		gp.lineTo((float)(tortue.corX-rayon-tortue.taille_crayon),
				(float)(tortue.corY-rayon-tortue.taille_crayon));
		gp.lineTo((float)(tortue.corX-rayon-tortue.taille_crayon),
				(float)(tortue.corY-rayon-tortue.taille_crayon));*/
		if (null==rec) rec=new Rectangle();
		rec.setRect(tortue.corX-rayon-tortue.getPenWidth(),
			tortue.corY-rayon-tortue.getPenWidth(),
			2*(rayon+tortue.getPenWidth()),2*(rayon+tortue.getPenWidth()));
		clip(rec);
		montrecacheTortue(); // on efface la tortue si elle st visible
	}
	protected void circle(double rayon) {
		montrecacheTortue();
		if (null==arc) arc=new Arc2D.Double();
		arc.setArcByCenter(tortue.corX,tortue.corY,rayon,
				0,360, Arc2D.OPEN);
		if (tortue.isPenReverse()) {
			g.setColor(couleurfond);
			g.setXORMode(tortue.couleurcrayon);
		} else {
			g.setColor(tortue.couleurcrayon);
			g.setPaintMode();
		}
		g.draw(arc);
		if (null==clipArc) clipArc=new Arc2D.Double();
		clipArc.setArcByCenter(tortue.corX,tortue.corY,
				rayon+2+tortue.getPenWidth(),0,360, Arc2D.OPEN);
		clip(clipArc);
		montrecacheTortue(); // on efface la tortue si elle st visible
	}

	protected void chargeimage(BufferedImage image) {
		if (tortue.isVisible())
			montrecacheTortue();
		g.setPaintMode();
		g.setClip((int) tortue.corX, (int) tortue.corY, image
				.getWidth(), image.getHeight());
		g.drawImage(image, null, (int) tortue.corX,
				(int) tortue.corY);
		g.setClip(0, 0, Config.imageWidth,Config.imageHeight);
		graph.setPaintMode();
		if (null==rec) rec=new Rectangle();
		rec.setRect(tortue.corX,tortue.corY,
				image.getWidth(),image.getHeight());
		clip(rec);
		if (classicMode) graph.setClip(0, 0, Config.imageWidth,Config.imageHeight);
		if (tortue.isVisible())
			montrecacheTortue();
	}

	private double trouve_longueur(double mini, double maxi, double oldx, double oldy) {
		// renvoie la longueur dont on peut encore avancer
		if (Math.abs(maxi - mini) < 0.5){
			return (mini);}
		else {
			double milieu = (mini + maxi) / 2;
			double nx = oldx + milieu * Math.cos(tortue.angle);
			double ny = oldy - milieu * Math.sin(tortue.angle);
			if (nx < 0 || nx > Config.imageWidth|| ny < 0 || ny > Config.imageHeight)
				return trouve_longueur(mini, milieu, oldx, oldy);
			else
				return trouve_longueur(milieu, maxi, oldx, oldy);
		}
	}

	private void trace_enroule(double arg, double oldx, double oldy) {
		boolean re = false;
		if (arg < 0) {
			re = true;
		}
		double diagonale=Math.sqrt(Math.pow(Config.imageWidth,2)+Math.pow(Config.imageHeight,2))+1;
		double longueur;
		if (re)
			longueur = trouve_longueur(0, -diagonale, oldx, oldy);
		else
			longueur = trouve_longueur(0, diagonale, oldx, oldy);
//		System.out.println(diagonale+" "+oldx+" "+oldy);
		while (Math.abs(longueur) < Math.abs(arg)) {
		//	System.out.println(Math.abs(longueur)+" "+Math.abs(arg));
			arg -= longueur;
			DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_CLASSIC;
			av(longueur);
			//System.out.println(Math.abs(longueur)+" "+Math.abs(arg));
			if (cadre.error)
				break; //permet d'interrompre avec le bouton stop
			DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_WRAP;
			if (Config.turtleSpeed != 0) {
				try {
					Thread.sleep(Config.turtleSpeed * 5);
				} catch (InterruptedException e) {
				}
			}
			if (tortue.isVisible())
				this.montrecacheTortue();
				if (re) tortue.cap=(tortue.cap+180)%360;
			if (tortue.corX > Config.imageWidth-1
					&& (tortue.cap < 180 && tortue.cap != 0)) {
				tortue.corX = 0;
				if (tortue.corY > Config.imageHeight-1
						&& (tortue.cap > 90 && tortue.cap < 270))
					tortue.corY = 0;
				else if (tortue.corY < 1
						&& (tortue.cap < 90 || tortue.cap > 270))
					tortue.corY = Config.imageHeight;
			} else if (tortue.corX < 1 && tortue.cap > 180) {
				tortue.corX = Config.imageWidth;
				if (tortue.corY > Config.imageHeight-1
						&& (tortue.cap > 90 && tortue.cap < 270))
					tortue.corY = 0;
				else if (tortue.corY < 1
						&& (tortue.cap < 90 || tortue.cap > 270))
					tortue.corY = Config.imageHeight;
			} else if (tortue.corY > Config.imageHeight-1)
				tortue.corY = 0;
			else if (tortue.corY < 1)
				tortue.corY = Config.imageHeight;
			if (re) tortue.cap=(tortue.cap+180)%360;
			if (tortue.isVisible())
				this.montrecacheTortue();
			if (re)
				longueur = trouve_longueur(0, -diagonale, tortue.corX,
						tortue.corY);
			else
				longueur = trouve_longueur(0, diagonale, tortue.corX,
						tortue.corY);
		}
		DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_CLASSIC;
		if (!cadre.error)
			av(arg);
		DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_WRAP;
	}

	private void trace_ferme(double oldx, double oldy, double arg) throws myException {
		boolean re = false;
		double longueur;
		double diagonale=Math.sqrt(Math.pow(Config.imageWidth,2)+Math.pow(Config.imageHeight,2))+1;
		if (arg < 0)
			re = true;
		if (re)
			longueur = trouve_longueur(0, -diagonale, oldx, oldy);
		else
			longueur = trouve_longueur(0, diagonale, oldx, oldy);
		if (Math.abs(longueur) < Math.abs(arg))
			throw new myException(cadre, Logo.messages
					.getString("erreur_sortie1")
					+ "\n"
					+ Logo.messages.getString("erreur_sortie2")
					+ Math.abs((int) (longueur)));
		else {
			DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_CLASSIC;
			av(arg);
			DrawGraphics.etat_fenetre = DrawGraphics.WINDOW_CLOSE;
		}
	}
	private void extractCoords(String liste,String prim)throws myException{
		coords[0]=1;
		coords[1]=1;
		StringTokenizer st = new StringTokenizer(liste);
		try {
			if (!st.hasMoreTokens())
				throw new myException(cadre, prim
						+ " " + Logo.messages.getString("n_aime_pas") + liste
						+ Logo.messages.getString("comme_parametre"));
			String li = st.nextToken().toString();
			if (li.equals("-")) {
				if (st.hasMoreTokens())
					li = st.nextToken().toString();
				coords[0] = -1;
			}
			coords[0] = coords[0] * Double.parseDouble(li) + Config.imageWidth/2;
			if (!st.hasMoreTokens())
				throw new myException(cadre, prim
						+ " " + Logo.messages.getString("n_aime_pas") + liste
						+ Logo.messages.getString("comme_parametre"));
			li = st.nextToken().toString();
			if (li.equals("-")) {
				if (st.hasMoreTokens())
					li = st.nextToken().toString();
				coords[1] = -1;
			}
			coords[1] = Config.imageHeight/2 - Double.parseDouble(li) * coords[1];
		} catch (NumberFormatException e) {
			throw new myException(cadre, prim
					+ " " + Logo.messages.getString("n_aime_pas") + liste
					+ Logo.messages.getString("comme_parametre"));
		}
		if (st.hasMoreTokens())
			throw new myException(cadre, prim
					+ " " + Logo.messages.getString("n_aime_pas") + liste
					+ Logo.messages.getString("comme_parametre"));

	}
	protected void point(String liste) throws myException {
		if (null==coords) coords=new double[2];		
		extractCoords(liste,Logo.messages.getString("point"));
		if (coords[0]>0 && coords[1]>0 && coords[0]<Config.imageHeight && coords[1] < Config.imageWidth) {
			if (tortue.isPenReverse()) {
				g.setColor(couleurfond);
				g.setXORMode(tortue.couleurcrayon);
				
			} else {
				g.setColor(tortue.couleurcrayon);
				g.setPaintMode();
			}	
			if (rec==null) rec=new Rectangle2D.Double();
			 // High quality
			if (Config.quality==Config.QUALITY_HIGH){
				double width=tortue.getPenWidth();
				rec.setRect(coords[0]-width,coords[1]-width,
						2*width,2*width);
			}
			// Normal or Low Quality
			else{
				// penWidth is 2k or 2k+1??
				int intWidth=(int)(2*tortue.getPenWidth()+0.5);
				if (intWidth%2==1){
					double width=tortue.getPenWidth()-0.5;
//					System.out.println(coords[0]+" "+coords[1]);
					rec.setRect(coords[0]-width,coords[1]-width,
							2*width+1,2*width+1);
				}
				else {
					double width=tortue.getPenWidth();
					rec.setRect(coords[0]-width,coords[1]-width,
							2*width,2*width);	
				}
			}
				if (Config.penShape==Config.PEN_SHAPE_SQUARE){
					g.fill(rec);
				}
				else if (Config.penShape==Config.PEN_SHAPE_OVAL){
					if (null==arc) arc=new Arc2D.Double();
					arc.setArcByCenter(coords[0],coords[1],0,0,360,Arc2D.OPEN);
					g.draw(arc);
				}
				clip(rec);
			}
	}

	protected void fpos(String liste) throws myException {
		if (null==coords) coords=new double[2];		
		extractCoords(liste,Logo.messages.getString("point"));
		oldx = tortue.corX;
		oldy = tortue.corY;
		montrecacheTortue();
		tortue.corX = coords[0];
		tortue.corY = coords[1];
		if (tortue.isPenDown()) {
			if (tortue.isPenReverse()) {
				g.setColor(couleurfond);
				g.setXORMode(tortue.couleurcrayon);
			} else {
				g.setColor(tortue.couleurcrayon);
				g.setPaintMode();
			}
			if (null==line) line=new Line2D.Double();
			if (null==gp) gp=new GeneralPath();
			else gp.reset();
			if (oldx < tortue.corX){
				x1=oldx;y1=oldy;x2=tortue.corX;y2=tortue.corY;
			}
			if (oldx>tortue.corX){
				x2=oldx;y2=oldy;x1=tortue.corX;y1=tortue.corY;
			}
			else if (oldx==tortue.corX){
				if (oldy<tortue.corY){
					x2=oldx;y2=oldy;x1=tortue.corX;y1=tortue.corY;						
					}
				else{ 
					x1=oldx;y1=oldy;x2=tortue.corX;y2=tortue.corY;
				}
			}
			line.setLine(x1,y1,x2,y2);
			// perpendicular vector
			nx=y1-y2;
			ny=x2-x1;
			length=Math.sqrt(nx*nx+ny*ny);
			factor=(1+tortue.getPenWidth())/length;
			vx=x2-x1;
			vy=y2-y1;
			gp.moveTo((float)(x1-vx*factor-nx*factor),
					(float)(y1-vy*factor-ny*factor));
			gp.lineTo((float)(x1-vx*factor+nx*factor),
					(float)(y1-vy*factor+ny*factor));		
			gp.lineTo((float)(x2+vx*factor+nx*factor),
					(float)(y2+vy*factor+ny*factor));
			gp.lineTo((float)(x2+vx*factor-nx*factor),
					(float)(y2+vy*factor-ny*factor));
			gp.lineTo((float)(x1-vx*factor-nx*factor),
					(float)(y1-vy*factor-ny*factor));
			g.draw(line);
			clip(gp);
		}
		montrecacheTortue();
	}
	protected void ct_mt() {
		if (null == tortue.tort) {
			g.setXORMode(couleurfond);
			g.setColor(tortue.couleurcrayon);
			tortue.coord();
			BasicStroke crayon_actuel = (BasicStroke) g.getStroke();
			if (crayon_actuel.getLineWidth() == 1)
				g.draw(tortue.triangle);
			else {
				g.setStroke(crayon_triangle);
				g.draw(tortue.triangle);
				g.setStroke(crayon_actuel);
			}
		} else {
			g.setXORMode(couleurfond);
			double angle = Math.PI / 2 - tortue.angle;
			float x = (float) (tortue.corX * Math.cos(angle) + tortue.corY
					* Math.sin(angle));
			float y = (float) (-tortue.corX * Math.sin(angle) + tortue.corY
					* Math.cos(angle));
			g.rotate(angle);
			g.drawImage(tortue.tort, (int) x - tortue.largeur / 2,
					(int) y - tortue.hauteur / 2, cadre.getArdoise());
			g.rotate(-angle);
		}
		if (null==rec) rec=new Rectangle();
		rec.setRect(tortue.corX - tortue.gabarit,
				tortue.corY - tortue.gabarit,
				tortue.gabarit * 2,
				tortue.gabarit * 2);
		clip(rec);
	
/*		clip((int) (tortue.corX - tortue.gabarit),
				(int) (tortue.corY - tortue.gabarit),
				tortue.gabarit * 2, tortue.gabarit * 2);*/
	}

	protected void montrecacheTortue() {
		g.setColor(couleurfond);
		for (int i = 0; i < tortues_visibles.size(); i++) {
			int id = Integer.parseInt(String.valueOf(tortues_visibles
					.get(i)));

			if (null == tortues[id].tort) {
				g.setXORMode(couleurfond);
				g.setColor(tortues[id].couleurmodedessin);
				tortues[id].coord();
				BasicStroke crayon_actuel = (BasicStroke) g.getStroke();
				if (crayon_actuel.getLineWidth() == 1)
					g.draw(tortues[id].triangle);
				else {
					g.setStroke(crayon_triangle);
					g.draw(tortues[id].triangle);
					g.setStroke(crayon_actuel);
				}
			} else {
				g.setXORMode(couleurfond);
				double angle = Math.PI / 2 - tortues[id].angle;
				float x = (float) (tortues[id].corX * Math.cos(angle) + tortues[id].corY
						* Math.sin(angle));
				float y = (float) (-tortues[id].corX * Math.sin(angle) + tortues[id].corY
						* Math.cos(angle));
				g.rotate(angle);
				g.drawImage(tortues[id].tort, (int) x
						- tortues[id].largeur / 2, (int) y
						- tortues[id].hauteur / 2, cadre.getArdoise());
				g.rotate(-angle);
			}
			if (null==rec) rec=new Rectangle();
			rec.setRect(tortues[id].corX - tortues[id].gabarit,
					tortues[id].corY - tortues[id].gabarit,
					tortues[id].gabarit * 2,
					tortues[id].gabarit * 2);
			clip(rec);
		}
	}

	protected void td(double arg) {
//		System.out.println(tortue.angle);
		if (tortue.isVisible())
			montrecacheTortue();
		tortue.cap = ((tortue.cap + arg) % 360 + 360) % 360;
		tortue.angle = Math.toRadians(90 - tortue.cap);
		if (tortue.isVisible())
			montrecacheTortue();
		Interprete.operande = false;
	}
	protected void videecran() {
//		System.out.println(tortue.corX+" "+tortue.corY);
		g.setPaintMode();
		g.setColor(couleurfond);
		g.fillRect(0, 0, Config.imageWidth,Config.imageHeight);
		// Réinitialiser les tortues
		if (null == tortues[0])
			tortues[0] = new Turtle(cadre);
		tortue = tortues[0]; //la tortue active sera à présent la
		// numéro 0
		tortue.id = 0;
		for (int i = 1; i < tortues.length; i++) { //On élimine les
			// autres tortues
			tortues[i] = null;
		}
		tortues_visibles.removeAllElements();
//		System.out.println(tortue.corX+" "+tortue.corY);
		tortues_visibles.push("0");
		g.setColor(tortue.couleurcrayon);
		graph.setClip(0, 0, Config.imageWidth,Config.imageHeight);
		if (classicMode) {
					graph.drawImage(Ardoise.dessin, 0, 0, cadre.getArdoise());
		}
		tortue.init();
		tortue.setVisible(true);
		g.setStroke(new BasicStroke(1));
		montrecacheTortue();
	}

	protected void nettoie() {
		g.setPaintMode();
		g.setColor(couleurfond);
		g.fillRect(0, 0, Config.imageWidth,Config.imageHeight);
		/* Réinitialiser les tortues
		if (null == tortues[0])
			tortues[0] = new Tortue(cadre);
		tortue = tortues[0]; //la tortue active sera à présent la
		// numéro 0
		tortue.id = 0;
		for (int i = 1; i < tortues.length; i++) { //On élimine les
			// autres tortues
			tortues[i] = null;
		}
		tortues_visibles.removeAllElements();
		tortues_visibles.push("0");*/
		g.setColor(tortue.couleurcrayon);
		graph.setClip(0, 0, Config.imageWidth,Config.imageHeight);
		if (classicMode) graph.drawImage(Ardoise.dessin, 0, 0, cadre.getArdoise());
		if (tortue.isVisible())
			montrecacheTortue();
		else
			tortues_visibles=new Stack();
	}
	private int bornes_remplis_zone(int x, int y, int increment, int couleur_frontiere) {
		while (!meme_couleur(Ardoise.dessin.getRGB(x, y) ,couleur_frontiere)) {
			Ardoise.dessin.setRGB(x, y, couleur_frontiere);
			x = x + increment;
			if (!(x > 0 && x < Config.imageWidth-1))
				break;
		}
		return x - increment;
	}
	private boolean meme_couleur(int col1,int col2){
/*		if (Config.quality==Logo.QUALITY_HIGH){
			int rouge1 = (col1 >> 16) & 0xFF;
			int vert1 = (col1 >> 8) & 0xFF;
			int bleu1 = col1 & 0xFF;
			int rouge2 = (col2 >> 16) & 0xFF;
			int vert2 = (col2 >> 8) & 0xFF;
			int bleu2 = col2 & 0xFF;
			int tolerance=120;
			int diff_rouge=rouge1-rouge2;
			int diff_bleu=bleu1-bleu2;
			int diff_vert=vert1-vert2;
			boolean rouge;boolean vert; boolean bleu;
			if (rouge1>rouge2){
				if (rouge1-rouge2< 128 -rouge2/2) rouge=true;
				else rouge=false;
			}
			else{
				if (rouge2-rouge1<rouge2/2) rouge=true;
				else rouge=false;
			}
			if (vert1>vert2){
				if (vert1-vert2< 128 -vert2/2) vert=true;
				else vert=false;
			}
			else{
				if (vert2-vert1<vert2/2) vert=true;
				else vert=false;
			}
			if (bleu1>bleu2){
				if (bleu1-bleu2< 128 -bleu2/2) bleu=true;
				else bleu=false;
			}
			else{
				if (bleu2-bleu1<bleu2/2) bleu=true;
				else bleu=false;
			}

		return rouge&&bleu&&vert;	
//			if (Math.abs(rouge1-rouge2)<tolerance&&Math.abs(vert1-vert2)<tolerance&&Math.abs(bleu1-bleu2)<tolerance&&Math.abs(rouge1+bleu1+vert1-rouge2-bleu2-vert2)<450)
//			return true;
	//	else return false;
		}
		else{*/
			return (col1==col2);
		//}
	}
	protected void rempliszone() {
		montrecacheTortue();
		int x = (int) (tortue.corX + 0.5);
		int y = (int) (tortue.corY + 0.5);
		if (x > 0 & x < Config.imageWidth & y > 0 & y < Config.imageHeight) {
			int couleur_origine = Ardoise.dessin.getRGB(x, y);
			int couleur_frontiere = tortue.couleurcrayon.getRGB();
		//	System.out.println(couleur_origine+" " +couleur_frontiere);
			Stack pile_germes = new Stack();
			boolean couleurs_differentes = tortue.isPenDown()
					&& !meme_couleur(couleur_origine,couleur_frontiere);
			if (couleurs_differentes)
				pile_germes.push(new Point(x, y));
			while (!pile_germes.isEmpty()) {

				Point p = (Point) pile_germes.pop();
				int xgerme = p.x;
				int ygerme = p.y;
				int xmax = bornes_remplis_zone(xgerme, ygerme, 1,
						couleur_frontiere);
				int xmin = bornes_remplis_zone(xgerme - 1, ygerme, -1,
						couleur_frontiere);
				boolean ligne_dessus = false;
				boolean ligne_dessous = false;
				for (int i = xmin; i < xmax + 1; i++) {
					//on recherche les germes au dessus et au dessous
					if (ygerme > 0
							&& meme_couleur(Ardoise.dessin.getRGB(i, ygerme - 1) ,couleur_frontiere)) {
						if (ligne_dessus)
							pile_germes.push(new Point(i - 1, ygerme - 1));
						ligne_dessus = false;
					} else {
						ligne_dessus = true;
						if (i == xmax && ygerme > 0)
							pile_germes.push(new Point(xmax, ygerme - 1));
					}
					if (ygerme < Config.imageHeight-1
							&& meme_couleur(Ardoise.dessin.getRGB(i, ygerme + 1),couleur_frontiere)) {
						if (ligne_dessous)
							pile_germes.push(new Point(i - 1, ygerme + 1));
						ligne_dessous = false;
					} else {
						ligne_dessous = true;
						if (i == xmax && ygerme < Config.imageHeight-1)
							pile_germes.push(new Point(xmax, ygerme + 1));
					}
				}
			}
			if (classicMode){
				graph.setClip(0, 0,Config.imageWidth,Config.imageHeight);
				graph.drawImage(Ardoise.dessin, 0, 0, cadre);
			}
			montrecacheTortue();
		}
	}

	private int bornes_remplis(int x, int y, int increment, int couleur_crayon,
			int couleur_origine) {
		while (Ardoise.dessin.getRGB(x, y) == couleur_origine) {
			Ardoise.dessin.setRGB(x, y, couleur_crayon);
			x = x + increment;
			if (!(x > 0 && x < Config.imageWidth-1))
				break;
		}
		return x - increment;
	}

	protected void remplis() {
		montrecacheTortue();
		int x = (int) (tortue.corX + 0.5);
		int y = (int) (tortue.corY + 0.5);
		if (x > 0 & x < Config.imageWidth & y > 0 & y < Config.imageHeight) {
		int couleur_origine = Ardoise.dessin.getRGB(x, y);
		int couleur_crayon = tortue.couleurcrayon.getRGB();
		if (x > 0 & x < Config.imageWidth & y > 0 & y < Config.imageHeight) {
			Stack pile_germes = new Stack();
			boolean couleurs_differentes = tortue.isPenDown()
					&& !(couleur_origine == couleur_crayon);
			if (couleurs_differentes)
				pile_germes.push(new Point(x, y));
			while (!pile_germes.isEmpty()) {
				Point p = (Point) pile_germes.pop();
				int xgerme = p.x;
				int ygerme = p.y;
				//			System.out.println(xgerme+" "+ygerme);
				int xmax = bornes_remplis(xgerme, ygerme, 1, couleur_crayon,
						couleur_origine);
				int xmin = bornes_remplis(xgerme - 1, ygerme, -1,
						couleur_crayon, couleur_origine);
				//				System.out.println("xmax "+xmax+"xmin "+xmin);
				boolean ligne_dessus = false;
				boolean ligne_dessous = false;
				for (int i = xmin; i < xmax + 1; i++) {
					//on recherche les germes au dessus et au dessous
					if (ygerme > 0
							&& Ardoise.dessin.getRGB(i, ygerme - 1) != couleur_origine) {
						if (ligne_dessus)
							pile_germes.push(new Point(i - 1, ygerme - 1));
						ligne_dessus = false;
					} else {
						ligne_dessus = true;
						if (i == xmax && ygerme > 0)
							pile_germes.push(new Point(xmax, ygerme - 1));
					}
					if (ygerme < Config.imageHeight-1
							&& Ardoise.dessin.getRGB(i, ygerme + 1) != couleur_origine) {
						if (ligne_dessous)
							pile_germes.push(new Point(i - 1, ygerme + 1));
						ligne_dessous = false;
					} else {
						ligne_dessous = true;
						if (i == xmax && ygerme < Config.imageHeight-1)
							pile_germes.push(new Point(xmax, ygerme + 1));
					}
				}
			}
			if (classicMode){
				graph.setClip(0, 0, Config.imageWidth,Config.imageHeight);
				graph.drawImage(Ardoise.dessin, 0, 0, cadre);
			}
			montrecacheTortue();
		}
		}
	}
	protected void etiquette(String mot) {
		//	Graphics2D g = (Graphics2D) Ardoise.dessin.getGraphics();
		montrecacheTortue();
		double angle = Math.PI / 2 - tortue.angle;
		g.rotate(angle);
		g.setPaintMode();
		g.setColor(tortue.couleurcrayon);
		float x = (float) (tortue.corX * Math.cos(angle) + tortue.corY
				* Math.sin(angle));
		float y = (float) (-tortue.corX * Math.sin(angle) + tortue.corY
				* Math.cos(angle));
		g.setFont(Panel_Font.fontes[police_etiquette]
				.deriveFont((float) tortue.police));
		g.drawString(mot, x, y);
		g.rotate(-angle);		
		montrecacheTortue();
		if (classicMode) cadre.getArdoise().repaint();
	}
	// set Screen Color
	protected void fcfg(Color color) {
		Graphics g = Ardoise.dessin.getGraphics();
		g.setColor(color);
		g.fillRect(0, 0, Config.imageWidth,Config.imageHeight);
		couleurfond=color;
		videecran();
		cadre.getArdoise().repaint();
	}
	protected void fcc(Color color) {
		if (tortue.isVisible()&&null==tortue.tort) montrecacheTortue();
		tortue.couleurcrayon = color;
		tortue.couleurmodedessin = color;
		if (tortue.isVisible()&&null==tortue.tort) montrecacheTortue();
	}
	protected void rafraichis(){ //Lorsqu'on change la taille de la zone de dessin
       g=(Graphics2D)Ardoise.dessin.getGraphics();
		if (Config.quality==Config.QUALITY_HIGH) g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
       graph=cadre.getArdoise().getGraphics();
		
	}
	protected void rafraichis_animation(){
		graph.setClip(0,0,Config.imageWidth,Config.imageHeight);
		graph.drawImage(Ardoise.dessin, 0, 0, cadre);
	}
	protected void origine(){ // primitive origine
		try {
			fpos("0 0");
		} catch (myException e) {
		}
		if (tortue.isVisible())
			montrecacheTortue();
			tortue.cap = 0;
			tortue.angle = Math.PI / 2;
		if (tortue.isVisible())
			montrecacheTortue();
	}
	
	private void clip(Shape gp){
		Rectangle rec1=cadre.jScrollPane1.getViewport().getViewRect();
		boolean inter=gp.intersects(rec1);
		if (inter){
			if (classicMode){
				graph.setClip(gp);
				graph.drawImage(Ardoise.dessin, 0, 0, cadre);
			}
		}
	}
	public void setQuality(int id){
		if (id==Config.QUALITY_HIGH){
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
			g.setRenderingHint(RenderingHints.KEY_RENDERING,RenderingHints.VALUE_RENDER_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		}
		else if(id==Config.QUALITY_LOW){
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_OFF);
			g.setRenderingHint(RenderingHints.KEY_RENDERING,RenderingHints.VALUE_RENDER_SPEED);
			g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
		}
		else { //normal
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_DEFAULT);
			g.setRenderingHint(RenderingHints.KEY_RENDERING,RenderingHints.VALUE_RENDER_DEFAULT);
			g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT);
		}
	}
	// When user want to resize the drawing zone
	public void kill_Graphics(){
		g.dispose();
		graph.dispose();
	}
	public  void change_image_tortue(Application cadre, String chemin) {
		if (tortue.isVisible())
			montrecacheTortue();
		if (chemin.equals("tortue0.png")) {
			tortue.tort = null;
			tortue.largeur = 26;
			tortue.hauteur = 26;
		} else {
			//ON teste tout d'abord si le chemin est valide
			if (null == Utils.class.getResource(chemin))
				chemin = "tortue1.png";
			tortue.tort = Toolkit.getDefaultToolkit().getImage(
					Utils.class.getResource(chemin));
			MediaTracker tracker = new MediaTracker(cadre);
			tracker.addImage(tortue.tort, 0);
			try {
				tracker.waitForID(0);
			} catch (InterruptedException e1) {
			}
			double largeur_ecran = Toolkit.getDefaultToolkit().getScreenSize()
					.getWidth();
			// On fait attention à la résolution de l'utilisateur
			double facteur = largeur_ecran / 1024.0;

			if ((int) (facteur + 0.001) != 1) {
				tortue.largeur = tortue.tort.getWidth(cadre);
				tortue.hauteur = tortue.tort.getHeight(cadre);
				tortue.tort = tortue.tort.getScaledInstance(
						(int) (facteur * tortue.largeur),
						(int) (facteur * tortue.hauteur),
						Image.SCALE_SMOOTH);
				tracker = new MediaTracker(cadre);
				tracker.addImage(tortue.tort, 0);
				try {
					tracker.waitForID(0);
				} catch (InterruptedException e1) {
				}
			}
			tortue.largeur = tortue.tort.getWidth(cadre);
			tortue.hauteur = tortue.tort.getHeight(cadre);
		}
		tortue.gabarit = Math.max(tortue.hauteur,
				tortue.largeur);
		if (tortue.isVisible())
			montrecacheTortue();

	}
	// animation 
	protected void setAnimation(boolean predic) throws myException{
		if (predic) {
			classicMode=MODE_ANIMATION;
			cadre.getHistoryPanel().active_animation();
		}
		else {
			classicMode=MODE_CLASSIC;
			cadre.getHistoryPanel().stop_animation();
		}
	}
	
	
	protected void setFont(Font f){
		g.setFont(f);
	}
	protected Font getFont(){
		return g.getFont();
	}
	protected void setStroke(Stroke st){
		g.setStroke(st);
	}

	
	public void adjustmentValueChanged(AdjustmentEvent e) {
		// Si on déplace les srollbar
		//	Il faut indiquer à la tortue que
		//	la zone de dessin a changée

		if (null != kernel.primitive) {
			graph = cadre.getArdoise().getGraphics();
		}
	}
	protected Color getScreenBackground(){
		return couleurfond;
	}
	protected void setScreenBackground(Color c){
		couleurfond=c;
	}
	public void setNumberOfTurtles(int id){
		Config.maxTurtles = id;
		Turtle[] tampon = (Turtle[]) tortues.clone();
		tortues = new Turtle[id];
		int borne_sup=Math.min(tampon.length,tortues.length);
		for(int i=0;i<borne_sup;i++){
			tortues[i]=tampon[i];
		}
		for(int i=tortues_visibles.size()-1;i>-1;i--){
			int integer=Integer.parseInt(tortues_visibles.get(i).toString());
			if (integer>=id){
				tortues_visibles.remove(i);
			}
		}
	}
	protected void init(){
//		cadre.jScrollPane1.setPreferredSize(new java.awt.Dimension(Config.imageWidth,Config.imageHeight))	;	
		MediaTracker tracker;
	    g.setColor(Config.screencolor);
	    g.fillRect(0,0,Config.imageWidth,Config.imageHeight);
	    g.setColor(Config.screencolor);
	    if (0==Config.activeTurtle) {
	    	tortue.coord();
	        g.setXORMode(tortue.couleurcrayon);
	        g.draw(tortue.triangle);
	    }
	    else {
	    	tracker=new MediaTracker(cadre);
	        tracker.addImage(tortue.tort,0);
	        try{tracker.waitForID(0);}
	        catch(InterruptedException e){}
	        g.setXORMode(Config.screencolor);
	        if (tracker.checkID(0))  g.drawImage(tortue.tort, Config.imageWidth/2 - tortue.largeur / 2,
	                    Config.imageHeight/2 - tortue.hauteur/2, cadre.getArdoise());
	      }
	      graph.drawImage(Ardoise.dessin,0,0,cadre);
	}
	}